import logging

import numpy as np

from quick_pandas.sort import radix_argsort, radix_sort
from quick_pandas.wrappers.wrapper import Wrapper, FunctionWrapper


class ndarray_wrapper(Wrapper):
    __wrapped_class__ = np.ndarray
    __class__ = np.ndarray

    def __init__(self, *args, **kwargs):
        super(ndarray_wrapper, self).__init__(*args, **kwargs)
        self.__actual_class__ = ndarray_wrapper

    def sort(self, axis=-1, kind='quicksort', order=None):
        if kind == 'radixsort' and len(self.__wrapped_instance__.shape) == 1:
            return radix_sort(self.__wrapped_instance__)
        return self.__wrapped_instance__.sort(axis, kind, order)

    def argsort(self, axis=-1, kind='quicksort', order=None):
        if kind == 'radixsort' and len(self.__wrapped_instance__.shape) == 1:
            return radix_argsort(self.__wrapped_instance__)
        return self.__wrapped_instance__.sort(axis, kind, order)

    def __len__(self):
        return self.__wrapped_instance__.__len__()

    def __getitem__(self, item):
        return self.__wrapped_instance__[item]

    def __setitem__(self, key, value):
        self.__wrapped_instance__[key] = value

    def __eq__(self, other):
        return self.__wrapped_instance__.__eq__(other)

    def __gt__(self, other):
        return self.__wrapped_instance__.__gt__(other)

    def __ge__(self, other):
        return self.__wrapped_instance__.__ge__(other)

    def __lt__(self, other):
        return self.__wrapped_instance__.__lt__(other)

    def __le__(self, other):
        return self.__wrapped_instance__.__le__(other)


class NdarrayFunctionWrapper(FunctionWrapper):
    def __call__(self, *args, **kwargs):
        res = self.func(*args, **kwargs)
        if isinstance(res, ndarray_wrapper.__wrapped_class__):
            return ndarray_wrapper(__wrapped_instance__=res)
        else:
            logging.warning("The result %s is not an instance of numpy.ndarray")
            return res



import pandas

from quick_pandas import sort_api, config
from quick_pandas.wrappers.pandas_wrapper import pandas_core_sorting_nargsort


def patch_numpy():
    # np.ndarray = ndarray_wrapper
    # np.empty = NdarrayFunctionWrapper(np.empty)
    pass


def patch_pandas():
    pandas.core.sorting.nargsort = pandas_core_sorting_nargsort


def init(ascii_sort=False):
    config.ASCII_SORT = ascii_sort
    sort_api.init()


def patch_all():
    init()
    patch_numpy()
    patch_pandas()



import logging

from pandas.core.sorting import *

from quick_pandas import sort_api

logger = logging.getLogger('pandas_wrapper')


def argsort(array: np.ndarray, kind):
    if array.dtype.type in [np.int64, np.int32, np.str_, np.float32, np.float64] and kind == 'radixsort':
        return sort_api.radix_argsort(array)
    logger.warning('revert to quick sort')
    return array.argsort(kind='quicksort' if kind == 'radixsort' else kind)


def pandas_core_sorting_nargsort(items, kind='quicksort', ascending=True, na_position='last'):
    """
    This is intended to be a drop-in replacement for np.argsort which
    handles NaNs. It adds ascending and na_position parameters.
    GH #6399, #5231
    """

    # specially handle Categorical
    if is_categorical_dtype(items):
        if na_position not in {'first', 'last'}:
            raise ValueError('invalid na_position: {!r}'.format(na_position))

        mask = isna(items)
        cnt_null = mask.sum()
        sorted_idx = items.argsort(ascending=ascending, kind=kind)
        if ascending and na_position == 'last':
            # NaN is coded as -1 and is listed in front after sorting
            sorted_idx = np.roll(sorted_idx, -cnt_null)
        elif not ascending and na_position == 'first':
            # NaN is coded as -1 and is listed in the end after sorting
            sorted_idx = np.roll(sorted_idx, cnt_null)
        return sorted_idx

    with warnings.catch_warnings():
        # https://github.com/pandas-dev/pandas/issues/25439
        # can be removed once ExtensionArrays are properly handled by nargsort
        warnings.filterwarnings(
            "ignore", category=FutureWarning,
            message="Converting timezone-aware DatetimeArray to")
        items = np.asanyarray(items)
    idx = np.arange(len(items))
    mask = isna(items)
    non_nans = items[~mask]
    non_nan_idx = idx[~mask]
    nan_idx = np.nonzero(mask)[0]
    if not ascending:
        non_nans = non_nans[::-1]
        non_nan_idx = non_nan_idx[::-1]
    import time
    ts = time.time()
    # indexes = non_nans.argsort(kind=kind)
    indexes = argsort(non_nans, kind)
    print(type(non_nans))
    print('np argsort takes %fs on %s with type %s' % (time.time() - ts, non_nans.shape, non_nans.dtype))
    ts = time.time()
    indexer = non_nan_idx[indexes]
    print('getting indexer takes %fs on %s' % (time.time() - ts, non_nans.shape))
    if not ascending:
        indexer = indexer[::-1]
    # Finally, place the NaNs at the end or the beginning according to
    # na_position
    if na_position == 'last':
        indexer = np.concatenate([indexer, nan_idx])
    elif na_position == 'first':
        indexer = np.concatenate([nan_idx, indexer])
    else:
        raise ValueError('invalid na_position: {!r}'.format(na_position))
    return indexer



import numpy as np
from numba import njit

INSERTION_SORT_LIMIT = 64
RADIX_BITS = 8


@njit()
def radix_sort0(array: np.ndarray, array_offset: int, array_length: int):
    if array_length == 0:
        return
    if array_length <= INSERTION_SORT_LIMIT:
        for i in range(array_offset + 1, array_offset + array_length):
            val = array[i]
            j = i
            while j > array_offset and val < array[j - 1]:
                array[j] = array[j - 1]
                j -= 1
            array[j] = val
        return

    min_val = np.min(array[array_offset: array_offset + array_length])
    max_val = np.max(array[array_offset: array_offset + array_length])
    length = max_val - min_val
    value_bits = 64
    div = value_bits // 2
    while div > 0:
        if length >> div == 0:
            value_bits -= div
        else:
            length >>= div
        div >>= 1
    bits = min(value_bits, RADIX_BITS)
    last = bits == value_bits
    shift = value_bits - bits
    bin_length = 1 << bits
    bins = np.zeros(bin_length + 1, np.int32)
    for val in array[array_offset: array_offset + array_length]:
        bin_i = (val - min_val) >> shift
        bins[bin_i + 1] += 1
    for i in range(1, bin_length + 1):
        bins[i] += bins[i - 1]
    count = bins.copy()
    new_array = np.zeros_like(array[array_offset: array_offset + array_length])
    for val in array[array_offset: array_offset + array_length]:
        bin_i = (val - min_val) >> shift
        index = count[bin_i]
        count[bin_i] += 1
        new_array[index] = val
    array[array_offset:array_offset + array_length] = new_array
    if not last:
        for i in range(1, bin_length + 1):
            radix_sort0(array, array_offset + bins[i - 1], bins[i] - bins[i - 1])


def radix_sort(array: np.ndarray):
    radix_sort0(array, 0, len(array))


@njit()
def radix_argsort0(array: np.ndarray, indexes: np.ndarray, array_offset: int, array_length: int):
    if array_length == 0:
        return
    if array_length <= INSERTION_SORT_LIMIT:
        for i in range(array_offset + 1, array_offset + array_length):
            val = indexes[i]
            j = i
            while j > array_offset and array[val] < array[indexes[j - 1]]:
                indexes[j] = indexes[j - 1]
                j -= 1
            indexes[j] = val
        return
    min_val = array[indexes[array_offset]]
    max_val = array[indexes[array_offset]]
    for i in range(array_offset + 1, array_offset + array_length):
        index = indexes[i]
        if array[index] > max_val:
            max_val = array[index]
        if array[index] < min_val:
            min_val = array[index]
    length = max_val - min_val
    value_bits = array.itemsize * 8
    div = value_bits // 2
    while div > 0:
        if length >> div == 0:
            value_bits -= div
        else:
            length >>= div
        div >>= 1
    bits = min(value_bits, RADIX_BITS)
    last = bits == value_bits
    shift = value_bits - bits
    bin_length = 1 << bits
    bins = np.zeros(bin_length + 1, np.int32)
    for index in indexes[array_offset: array_offset + array_length]:
        val = array[index]
        bin_i = (val - min_val) >> shift
        bins[bin_i + 1] += 1
    for i in range(1, bin_length + 1):
        bins[i] += bins[i - 1]
    count = bins.copy()
    new_indexes = np.zeros_like(indexes[array_offset: array_offset + array_length])
    for val in indexes[array_offset: array_offset + array_length]:
        bin_i = (array[val] - min_val) >> shift
        index = count[bin_i]
        count[bin_i] += 1
        new_indexes[index] = val
    indexes[array_offset:array_offset + array_length] = new_indexes
    if not last:
        for i in range(1, bin_length + 1):
            radix_argsort0(array, indexes, array_offset + bins[i - 1], bins[i] - bins[i - 1])


@njit()
def array_cmp_lt(a, b, offset, length, unicode):
    if length == 0:
        return False
    if a[offset] < b[offset]:
        return True
    elif a[offset] > b[offset]:
        return False
    elif (not unicode or (offset % 4) == 0) and a[offset] == 0:
        if b[offset] != 0:
            return True
        else:
            return False
    return array_cmp_lt(a, b, offset + 1, length - 1, unicode)


@njit()
def radix_argsort0_str(array: np.ndarray, indexes: np.ndarray, array_offset: int, array_length: int, str_offset: int,
                       unicode: bool):
    if array_length == 0:
        return
    if str_offset >= len(array[0]):
        return
    if array_length < INSERTION_SORT_LIMIT:
        for i in range(array_offset + 1, array_offset + array_length):
            val = indexes[i]
            j = i
            while j > array_offset and array_cmp_lt(array[val], array[indexes[j - 1]], str_offset,
                                                    array.shape[1] - str_offset, unicode):
                indexes[j] = indexes[j - 1]
                j -= 1
            indexes[j] = val
        return

    min_val = array[indexes[array_offset]][str_offset]
    max_val = array[indexes[array_offset]][str_offset]
    for index in indexes[array_offset: array_offset + array_length]:
        val = array[index][str_offset]
        if val < min_val:
            min_val = val
        elif val > max_val:
            max_val = val
    if min_val == max_val:
        if max_val == 0 and (unicode or str_offset % 4) == 0:
            return
        else:
            return radix_argsort0_str(array, indexes, array_offset, array_length, str_offset + 1, unicode)
    length = max_val - min_val
    value_bits = 8
    div = value_bits // 2
    while div > 0:
        if length >> div == 0:
            value_bits -= div
        else:
            length >>= div
        div >>= 1
    bins = np.zeros((1 << value_bits) + 1, np.int32)
    for index in indexes[array_offset: array_offset + array_length]:
        val = array[index][str_offset]
        bin_i = val - min_val
        bins[bin_i + 1] += 1
    for i in range(1, len(bins)):
        bins[i] += bins[i - 1]
    count = bins.copy()
    new_indexes = np.zeros_like(indexes[array_offset: array_offset + array_length])
    for val in indexes[array_offset: array_offset + array_length]:
        bin_i = array[val][str_offset] - min_val
        index = count[bin_i]
        count[bin_i] += 1
        new_indexes[index] = val
    indexes[array_offset:array_offset + array_length] = new_indexes
    for i in range(1, len(bins)):
        if i == 1 and min_val == 0 and (unicode or str_offset % 4 == 0):
            continue
        radix_argsort0_str(array, indexes, array_offset + bins[i - 1], bins[i] - bins[i - 1], str_offset + 1, unicode)


@njit()
def convert_float64(item):
    if (item & np.uint64(0x8000000000000000)) == 0:
        return item ^ np.uint64(0x8000000000000000)
    return item ^ np.uint64(0xffffffffffffffff)


@njit()
def convert_float32(item):
    if (item & np.uint32(0x80000000)) == 0:
        return item ^ np.uint32(0x80000000)
    return item ^ np.uint32(0xffffffff)


@njit()
def radix_argsort0_float(array_float: np.ndarray, array: np.ndarray, indexes: np.ndarray, array_offset: int,
                         array_length: int, value_bits: int):
    if array_length <= 1 or value_bits == 0:
        return
    double = array.itemsize == 8
    if array_length <= INSERTION_SORT_LIMIT:
        for i in range(array_offset + 1, array_offset + array_length):
            index = indexes[i]
            j = i
            now = array_float[index]
            pre = array_float[indexes[j - 1]]
            while j > array_offset and (now < pre or (np.isnan(pre) and not np.isnan(now))):
                indexes[j] = indexes[j - 1]
                j -= 1
                pre = array_float[indexes[j - 1]]
            indexes[j] = index
        return
    value_bits_total = array.itemsize * 8
    value_mask = np.uint64(-1) >> np.uint64(value_bits_total - value_bits) \
        if double else np.uint32(-1) >> np.uint32(value_bits_total - value_bits)
    bits = min(value_bits, RADIX_BITS)
    shift = value_bits - bits
    bin_length = 1 << bits
    bins = np.zeros(bin_length + 1, np.int32)
    for index in indexes[array_offset: array_offset + array_length]:
        array_value = array[index]
        if double:
            array_value = convert_float64(array_value) & value_mask
        else:
            array_value = convert_float32(array_value) & value_mask
        bin_i = array_value >> np.uint32(shift)
        bins[bin_i + 1] += 1
    for i in range(1, bin_length + 1):
        bins[i] += bins[i - 1]
    count = bins.copy()
    new_indexes = np.zeros_like(indexes[array_offset: array_offset + array_length])
    for val in indexes[array_offset: array_offset + array_length]:
        array_value = array[val]
        if double:
            array_value = convert_float64(array_value) & value_mask
        else:
            array_value = convert_float32(array_value) & value_mask
        bin_i = array_value >> np.uint32(shift)
        index = count[bin_i]
        count[bin_i] += 1
        new_indexes[index] = val
    indexes[array_offset:array_offset + array_length] = new_indexes
    for i in range(1, bin_length + 1):
        radix_argsort0_float(array_float, array, indexes, array_offset + bins[i - 1], bins[i] - bins[i - 1], shift)


def radix_argsort(array: np.ndarray, unicode=True):
    indexes = np.arange(array.shape[0], dtype=int)
    if array.dtype in [int]:
        radix_argsort0(array, indexes, 0, len(array))
    elif array.dtype.type in [np.str_]:
        radix_argsort0_str(array.view(np.uint8 if unicode else np.uint32)
                           .reshape(-1, array.itemsize if unicode else array.itemsize // 4), indexes, 0,
                           len(array), 0, unicode=unicode)
    elif array.dtype.type in [np.float32, np.float64]:
        radix_argsort0_float(array, array.view(np.uint64 if array.itemsize == 8 else np.uint32),
                             indexes, 0, len(array), array.itemsize * 8)
    else:
        raise Exception('unsupported data type %s' % array.dtype)
    return indexes



import logging
import time

import numpy as np

from quick_pandas.sort_api import radix_sort, radix_argsort


def radix_sort_str_benchmark():
    array_range = 10000
    array_length = 10000000
    array = np.random.randint(0, array_range, (array_length,))
    array = array.astype(str)
    radix_argsort(np.zeros((1,), dtype=str))
    radix_argsort(np.zeros((1,), dtype=str))
    for i in range(3):
        ts = time.time()
        np.argsort(array, kind='quicksort')
        logging.warning("[str] quick argsort causes %fs" % (time.time() - ts))
        ts = time.time()
        radix_argsort(array)
        logging.warning("[str] radix argsort(unicode=True) causes %fs" % (time.time() - ts))
        ts = time.time()
        radix_argsort(array, unicode=False)
        logging.warning("[str] radix argsort(unicode=False) causes %fs" % (time.time() - ts))


def radix_sort_float_benchmark():
    array_range = 10000
    array_length = 10000000
    array = np.random.rand(array_length)
    array -= 0.5
    array *= array_range
    for dtype in [np.float64, np.float32]:
        array = array.astype(dtype)
        np.argsort(np.zeros((1,), dtype=dtype))
        radix_argsort(np.zeros((1,), dtype=dtype))
        for i in range(3):
            ts = time.time()
            np.argsort(array, kind='quicksort')
            logging.warning("[%s] quick argsort causes %fs" % (dtype, time.time() - ts))
            ts = time.time()
            radix_argsort(array)
            logging.warning("[%s] radix argsort causes %fs" % (dtype, time.time() - ts))


def radix_sort_int_benchmark():
    array_range = 10000
    array_length = 100000000
    array = np.random.randint(-array_range, array_range, (array_length,))
    radix_sort(np.zeros((1,), dtype=int))
    radix_argsort(np.zeros((1,), dtype=int))
    for i in range(3):
        ts = time.time()
        np.argsort(array, kind='quicksort')
        logging.warning("[int] quick argsort causes %fs" % (time.time() - ts))
        ts = time.time()
        radix_argsort(array)
        logging.warning("[int] radix argsort causes %fs" % (time.time() - ts))
        ts = time.time()
        radix_sort(array)
        logging.warning("[int] radix sort causes %fs" % (time.time() - ts))


if __name__ == '__main__':
    radix_sort_int_benchmark()
    radix_sort_str_benchmark()
    radix_sort_float_benchmark()



import time

import numpy as np
import pandas as pd

from quick_pandas import monkey

monkey.patch_all()


def sort():
    data1 = np.random.randint(0, 10000, 10000000)
    data2 = data1.copy()
    df1 = pd.DataFrame(data=data1)
    df2 = pd.DataFrame(data=data2)
    ts = time.time()
    df1.sort_values(kind='quicksort', by=0)
    print('quick sort takes %fs' % (time.time() - ts))
    ts = time.time()
    df2.sort_values(kind='radixsort', by=0)
    print('radix sort takes %fs' % (time.time() - ts))


if __name__ == '__main__':
    sort()



class Wrapper(object):
    __wrapped_class__ = None

    def __init__(self, *args, **kwargs):
        if not self.__wrapped_class__:
            raise NotImplemented('__wrapped_class__ is not specified')
        __wrapped_instance__ = kwargs.pop('__wrapped_instance__', None)
        if __wrapped_instance__ is not None:
            if not isinstance(__wrapped_instance__, self.__wrapped_class__):
                raise Exception('wrapped instance %s is not an instance of wrapped class %s' % (
                    __wrapped_instance__, self.__wrapped_class__))
            self.__wrapped_instance__ = __wrapped_instance__
            self.__wrapped_class__ = __wrapped_instance__.__class__
        else:
            self.__wrapped_instance__ = self.__wrapped_class__(*args, **kwargs)

    def __getattr__(self, item):
        orig_attr = self.__wrapped_instance__.__getattribute__(item)
        if callable(orig_attr):
            def hooked(*args, **kwargs):
                result = orig_attr(*args, **kwargs)
                if result == self.__wrapped_instance__:
                    return self
                return result

            return hooked
        else:
            return orig_attr


class FunctionWrapper(object):
    def __init__(self, func):
        self.func = func

    def __getattr__(self, item):
        return self.func.__getattribute__(item)

    def __call__(self, *args, **kwargs):
        raise NotImplemented('a function wrapper should hook the original function by overwriting __call__')



import numpy as np

from quick_pandas import sort, config
from quick_pandas.wrappers.numpy_wrapper import ndarray_wrapper


def radix_argsort(array, unicode=None):
    if getattr(array, '__actual_class__', None) == ndarray_wrapper:
        array = array.__wrapped_instance__
    if unicode is None:
        unicode = not config.ASCII_SORT
    return sort.radix_argsort(array, unicode=unicode)


def radix_sort(array):
    if getattr(array, '__actual_class__', None) == ndarray_wrapper:
        array = array.__wrapped_instance__
    return sort.radix_sort(array)


def init():
    radix_argsort(np.array([1], dtype=int))
    radix_sort(np.array([1], dtype=int))
